// ribocore
// code by jph

#include "opengl.h"

#define TIC 0.0001
#define MUL 100
#define PVD 50
double dt;

array<float> rndu("data/rndu.tab");
#include "ani.hpp"
#include "aniProbe.cpp"

animated* ani1;
animated* ani2;
matrix<double> normal(4,1);
matrix<double> contac(4,1);
matrix<double> ini1Mrot;
matrix<double> ini2Mrot;
matrix<double> iniSMrot;
matrix<double> ini1Mkin;
matrix<double> ini2Mkin;
matrix<double> iniSMkin;
double max_penetration=0;

//defines body and collision characteristics
void configuration(int type) {  
  int c1;
  matrix<double> pos1(4,1);
  matrix<double> pos2(4,1);
  matrix<double> tst1(3,1,"fill");
  matrix<double> tst2(3,1,"fill");	
  matrix<double> Ome(4,4);
  switch (type) {
  case 1: // billiard balls
	  tst1+=1;
	  tst2.data[0]=-.3;
	  tst2.data[1]= .8;
	  tst2.data[2]=1.2;
	  ani1->pos=tst1.glDeriv();
    ani2->pos=tst2.glDeriv();
	  ani1->pos.exp();
	  ani2->pos.exp();	  	  
  	ani1->pos.data[12+0]=2;
  	ani1->pos.data[12+1]=.5;
  	ani1->vel.data[0]=-2;
  	ani1->setEllipsoid(.7,.7,.7);
  	ani2->setEllipsoid(1,1,1);
  	ani2->index=1;
    for (c1=0;c1<4;c1++) {
      pos1.data[c1]=ani1->pos.data[12+c1];
      pos2.data[c1]=ani2->pos.data[12+c1];
    }
    pos1-=pos2;
    pos1*=0.5;
    pos2+=pos1;
    normal=pos1;
    normal.normalize(2);
    contac=pos2;    
    break;
  case 2: // baseball
  	ani1->pos.data[12]=1.8;
  	ani1->pos.data[13]=.3;  	
  	ani1->vel.data[1]=-1.5;
  	ani2->ome.data[2]=1.5;
  	ani1->setEllipsoid(.3,.3,.2);
  	ani2->setEllipsoid(2.2,.3,.3);
  	ani2->index=1;
    normal.data[1]=1;    
    normal.normalize(2);
    contac.data[0]=1.8;
    contac.data[1]=.3;      
    break;  
  case 3: // stuntmen
    tst1=0;
    tst1.data[2]=.5; Ome=tst1.glDeriv(); Ome.exp();
    ani1->pos=ani1->pos*Ome;
    tst1.data[2]=-.4; Ome=tst1.glDeriv(); Ome.exp();
    ani2->pos=ani2->pos*Ome;
  	ani1->pos.data[12]=1.9;
  	ani1->pos.data[14]=.5;
  	ani1->vel.data[0]=-2;
  	ani1->vel.data[1]=.3;
  	ani1->ome.data[2]=1;
  	ani1->setEllipsoid(.5,1.8,.4);
  	ani2->setEllipsoid(.3,1.9,.6);
  	ani2->index=1;
    normal.data[0]=1;
    contac.data[0]=.7;
    contac.data[1]=1.3;
    contac.data[2]=.25;
    break;
  }
  contac.data[3]=1;
}

double complam(void) {
  double nom=0;
  double den=0;
  matrix<double> q1=ani1->Iine()*ani1->pos.glInverse()*(ani1->ini_lev&normal);
  matrix<double> q2=ani2->Iine()*ani2->pos.glInverse()*(ani2->ini_lev&normal);
  nom-=ani1->vel|normal;
  nom+=ani2->vel|normal;
  nom-=ani1->ine*q1|ani1->ome;
  nom+=ani2->ine*q2|ani2->ome;
  den+=1/ani1->mass;
  den+=1/ani2->mass;
  den+=ani1->ine*q1|q1;
  den+=ani2->ine*q2|q2;
  return 2*nom/den;
}

void main_(int argc,char** argv) {
	int c1,c2,cnt=0,ofs;
	glClearColor(.2,.2,.2,0);
	
	ani1=new aniProbe(1,0,0);
	ani2=new aniProbe(0,0,1);
	
	configuration(2);
	
	//------------------------------------------------
	//auto adjust
	//center of total mass at 0	
	matrix<double> cen(4,1);	
	cen+=ani1->loc()*ani1->mass;
	cen+=ani2->loc()*ani2->mass;	
	cen*=1/cen.data[3];	
	for (c1=0;c1<3;c1++) {
	  ani1->pos.data[12+c1]-=cen.data[c1];
	  ani2->pos.data[12+c1]-=cen.data[c1];
	  contac.data[c1]-=cen.data[c1];
  }
	//total impulse at 0
	cen*=0;
	cen+=ani1->vel*ani1->mass;
	cen+=ani2->vel*ani2->mass;
	cen*=1/(ani1->mass+ani2->mass);
  ani1->vel-=cen;
  ani2->vel-=cen;
  ani1->lev=ani1->pos.glInverse()*contac;
  ani2->lev=ani2->pos.glInverse()*contac;
  ani1->ini_lev=contac-ani1->loc();
  ani2->ini_lev=contac-ani2->loc();
  ani1->freeze();
  ani2->freeze();
  
  ini1Mrot=ani1->Mrot();
  ini2Mrot=ani2->Mrot();
  iniSMrot=ini1Mrot+ini2Mrot;
  ini1Mkin=ani1->Mkin();
  ini2Mkin=ani2->Mkin();
  iniSMkin=ini1Mkin+ini2Mkin;
  
  dt=-TIC;
  for (c1=0;c1<PVD*50;c1++) {
    ani1->motion();
    ani2->motion();    
  }
  dt=TIC;  
  cout << "computed lambda   \t" << complam() << endl;
};

bool hasCollided=false;
void collide(void) {
  int c1;
  char glt[80];
  matrix<double> pnt1=ani1->pos*ani1->lev;
  matrix<double> pnt2=ani2->pos*ani2->lev;  
  matrix<double> diff=pnt2-pnt1;
  matrix<double> tmp;
  double pen=normal|diff;
  double lam;
  if (pen>max_penetration)
    max_penetration=pen;  
  if (pen>0) {
    hasCollided=true;
    matrix<double> mun=normal*pen*MUL;    
    ani1->acc+=mun*(1/ani1->mass);
    ani2->acc-=mun*(1/ani2->mass);
    matrix<double> rl1=contac-ani1->loc();
    matrix<double> rl2=contac-ani2->loc();    
    ani1->tor+=rl1&mun;
    ani2->tor-=rl2&mun;
    sprintf(glt,"contact");
  	linked_label.append(new glLabelText(frame_w-49,frame_h-11,glt,fnt0));
  } else {
    if (hasCollided) {
      double err;
      tmp=(ani1->ini_vel-ani1->vel)*(-ani1->mass);
      lam=tmp|normal;
      cout << "\t error " << (tmp-normal*lam).norm(2) << endl;
      tmp=(ani2->ini_vel-ani2->vel)*(+ani2->mass);
      lam=tmp|normal;
      cout << "\t error " << (tmp-normal*lam).norm(2) << endl;
      tmp=ani1->pos*ani1->ine*ani1->ome-ani1->ini_ang;
      err=(tmp-(ani1->ini_lev&(normal*lam))).norm(2);
      if (ABS(err)<.0001) err=0;
      cout << "\t error " << err << endl;
      tmp=ani2->pos*ani2->ine*ani2->ome-ani2->ini_ang;
      err=(tmp+(ani2->ini_lev&(normal*lam))).norm(2);
      if (ABS(err)<.0001) err=0;
      cout << "\t error " << err << endl;
      cout << "simulated lambda\t" << lam << endl;
      hasCollided=false;
    }    
  }
}

bool hasReflected=false;
void reflect(void) {
  int c1;
  char glt[80];  
  matrix<double> pnt1=ani1->pos*ani1->lev;
  matrix<double> pnt2=ani2->pos*ani2->lev;  
  matrix<double> diff=pnt2-pnt1;
  double pen=normal|diff;
  if (pen>max_penetration)
    max_penetration=pen;  
  if (pen>0&&!hasReflected) {
    hasReflected=true;
    double lam=complam(); // *.9  for friction
    ani1->vel+=normal*(lam/ani1->mass);
    ani2->vel-=normal*(lam/ani2->mass);
    ani1->ome+=ani1->Iine()*ani1->pos.glInverse()*(ani1->ini_lev&normal)*lam;
    ani2->ome-=ani2->Iine()*ani2->pos.glInverse()*(ani2->ini_lev&normal)*lam;
  }
}

void display_(void) {
	int c1,c2,cnt=0;
	char glt[80];		
#ifndef RELEASE_MODE
  sprintf(glt,"rot %5.2f %5.2f %5.2f",ini1Mrot.data[0],ini1Mrot.data[1],ini1Mrot.data[2]);
	linked_label.append(new glLabelText(0,frame_h-11-11*0,glt,fnt0));
  sprintf(glt,"rot %5.2f %5.2f %5.2f",ini2Mrot.data[0],ini2Mrot.data[1],ini2Mrot.data[2]);
	linked_label.append(new glLabelText(0,frame_h-11-11*1,glt,fnt0));
  sprintf(glt,"rot %5.2f %5.2f %5.2f",iniSMrot.data[0],iniSMrot.data[1],iniSMrot.data[2]);
	linked_label.append(new glLabelText(0,frame_h-11-11*2,glt,fnt0));
  sprintf(glt,"kin %5.2f %5.2f %5.2f",ini1Mkin.data[0],ini1Mkin.data[1],ini1Mkin.data[2]);
	linked_label.append(new glLabelText(0,frame_h-44-11*0,glt,fnt0));
  sprintf(glt,"kin %5.2f %5.2f %5.2f",ini2Mkin.data[0],ini2Mkin.data[1],ini2Mkin.data[2]);
	linked_label.append(new glLabelText(0,frame_h-44-11*1,glt,fnt0));
  sprintf(glt,"kin %5.2f %5.2f %5.2f",iniSMkin.data[0],iniSMkin.data[1],iniSMkin.data[2]);
	linked_label.append(new glLabelText(0,frame_h-44-11*2,glt,fnt0));
#endif	
	matrix<float> d(4,1,""); d.data[1]=1; d.data[2]=1;
	glLightfv(GL_LIGHT0,GL_POSITION,d.data);
  glTranslated(0,0,-5);
  glDisable(GL_DEPTH_TEST);
  glColor3f(.3,.3,.3);
  glLineWidth(1);
  glBegin(GL_LINES);
    for (float x=-10;x<10;x+=1) {
    	glVertex3f(-10,x,0);
    	glVertex3f( 10,x,0);
    	glVertex3f(x,-10,0);
    	glVertex3f(x, 10,0);
    }
  glEnd();  	
  glEnable(GL_DEPTH_TEST);  
  ani1->draw();  
  ani2->draw();
  matrix<double> any(4,1,"");
  for (c1=0;c1<PVD;c1++) {
    ani1->motion();
    ani2->motion();    
    if (0)
      collide();
    else
      reflect();
  }
  
#ifndef RELEASE_MODE
  matrix<double> SMrot=ani1->Mrot()+ani2->Mrot();
  sprintf(glt,"rot %5.2f %5.2f %5.2f",SMrot.data[0],SMrot.data[1],SMrot.data[2]);
	linked_label.append(new glLabelText(200,frame_h-11-11*2,glt,fnt0));  
  matrix<double> SMkin=ani1->Mkin()+ani2->Mkin();
  sprintf(glt,"kin %5.2f %5.2f %5.2f",SMkin.data[0],SMkin.data[1],SMkin.data[2]);
	linked_label.append(new glLabelText(200,frame_h-44-11*2,glt,fnt0));  
  double Eani1=ani1->Ekin()+ani1->Erot();
  double Eani2=ani2->Ekin()+ani2->Erot();
  sprintf(glt,"Enew %5.2f %5.2f %5.2f",Eani1,Eani2,Eani1+Eani2);
	linked_label.append(new glLabelText(0,22,glt,fnt0));
	sprintf(glt,"Eold %5.2f %5.2f %5.2f",ani1->esum,ani2->esum,ani1->esum+ani2->esum);
	linked_label.append(new glLabelText(0,11,glt,fnt0));
#endif	

	glNormal3f(0,1,0);
  float u=0,w=5;  
  matrix<double> nme=contac+normal;
  glDisable(GL_DEPTH_TEST);		
	glColor3f(1,1,1);
	glLineWidth(2);
	glBegin(GL_LINES);
		glVertex3dv(contac.data);
		glVertex3dv(nme.data);
  glEnd();    
	glPointSize(4);
	glBegin(GL_POINTS);
		glVertex3dv(contac.data);
  glEnd();
  glEnable(GL_DEPTH_TEST);  
};